import logging
from pathlib import (
    Path,
)

import aiohttp_jinja2
import jinja2
from aiohttp import (
    web,
)
from aiohttp_middlewares import (
    cors_middleware,
)
from aiomisc.service.aiohttp import (
    AIOHTTPService,
)
from sqlalchemy import (
    create_engine,
)

from .config import (
    ApiGatewayConfig,
)
from .database.models import (
    Base,
)
from .handler import (
    AdminHandler,
    authentication,
    authentication_default,
    login_default,
    orchestrate,
)

logger = logging.getLogger(__name__)


class ApiGatewayRestService(AIOHTTPService):
    def __init__(self, address: str, port: int, config: ApiGatewayConfig):
        self.config = config
        self.engine = None
        super().__init__(address, port)

    async def create_application(self) -> web.Application:
        middlewares = list()
        if self.config.rest.cors.enabled:
            middlewares = [cors_middleware(allow_all=True)]

        app = web.Application(middlewares=middlewares)

        app["config"] = self.config

        self.engine = await self.create_engine()
        await self.create_database()

        app["db_engine"] = self.engine

        auth = self.config.rest.auth
        if auth is not None and auth.enabled:
            app.router.add_route("*", "/auth", authentication_default)
            app.router.add_route("*", "/auth/login", login_default)
            for service in auth.services:
                app.router.add_route("*", f"/auth/{service.name}", authentication)
                app.router.add_route("POST", f"/auth/{service.name}/login", authentication)

        app.router.add_route("POST", "/admin/login", AdminHandler.login)
        app.router.add_route("GET", "/admin/endpoints", AdminHandler.get_endpoints)
        app.router.add_route("GET", "/admin/rules", AdminHandler.get_rules)
        app.router.add_route("POST", "/admin/rules", AdminHandler.create_rule)
        app.router.add_route("PATCH", "/admin/rules/{id}", AdminHandler.update_rule)
        app.router.add_route("DELETE", "/admin/rules/{id}", AdminHandler.delete_rule)

        app.router.add_route("GET", "/admin/roles", AdminHandler.get_roles)
        app.router.add_route("POST", "/admin/autz-rules", AdminHandler.create_autz_rule)
        app.router.add_route("GET", "/admin/autz-rules", AdminHandler.get_autz_rules)
        app.router.add_route("PATCH", "/admin/autz-rules/{id}", AdminHandler.update_autz_rule)
        app.router.add_route("DELETE", "/admin/autz-rules/{id}", AdminHandler.delete_autz_rule)

        # Administration routes
        path = Path(__file__).parent
        aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(f"{path}/backend/templates"))
        app.router.add_route("*", "/administration{path:.*}", self.handler)
        # app.router.add_route("GET", "/administration/{filename:.*}", self._serve_files)

        app.router.add_route("*", "/{endpoint:.*}", orchestrate)

        return app

    async def create_engine(self):
        DATABASE_URI = (
            f"postgresql+psycopg2://{self.config.database.user}:{self.config.database.password}@"
            f"{self.config.database.host}:{self.config.database.port}/{self.config.database.dbname}"
        )

        return create_engine(DATABASE_URI)

    async def create_database(self):
        Base.metadata.create_all(self.engine)

    @aiohttp_jinja2.template("tmpl.jinja2")
    async def handler(self, request):  # pragma: no cover
        try:
            path = Path(__file__).parent
            filename = Path(request.match_info["path"].replace("/", "", 1))
            filepath = path.joinpath("backend", "admin", filename).resolve()

            if filepath.is_file():
                return await self._get_file(filepath)
        except Exception:
            pass
        response = aiohttp_jinja2.render_template("index.html", request, {})
        response.headers["Content-Language"] = "es"
        return response

    @staticmethod
    async def _get_file(file_path) -> web.FileResponse:  # pragma: no cover
        try:
            return web.FileResponse(path=file_path, status=200)
        except (ValueError, FileNotFoundError) as error:
            # relatively safe
            raise web.HTTPNotFound() from error
        except web.HTTPForbidden:
            raise
        except Exception as error:
            raise web.HTTPNotFound() from error

    """
    async def _serve_files(self, request: web.Request) -> web.FileResponse:
        rel_url = request.match_info["filename"]

        path = Path(Path.cwd())
        self._directory = path.resolve()

        try:
            filename = Path(rel_url)
            if filename.anchor:
                raise web.HTTPForbidden()
            filepath = self._directory.joinpath("minos", "api_gateway", "rest", "backend", "admin", filename).resolve()

            # p = filepath.relative_to(self._directory)
            return web.FileResponse(path=filepath, status=200)

        except (ValueError, FileNotFoundError) as error:
            # relatively safe
            raise web.HTTPNotFound() from error
        except web.HTTPForbidden:
            raise
        except Exception as error:
            # perm error or other kind!
            request.app.logger.exception(error)
            raise web.HTTPNotFound() from error
    """



from __future__ import (
    annotations,
)

import abc
import collections
import os
import typing as t
from distutils import (
    util,
)
from pathlib import (
    Path,
)
from typing import (
    Any,
)

import yaml

from .exceptions import (
    ApiGatewayConfigException,
)

REST = collections.namedtuple("Rest", "host port cors auth admin")
DISCOVERY = collections.namedtuple("Discovery", "host port")
CORS = collections.namedtuple("Cors", "enabled")
AUTH_SERVICE = collections.namedtuple("AuthService", "name")
REST_ADMIN = collections.namedtuple("RestAdmin", "username password")
DATABASE = collections.namedtuple("Database", "dbname user password host port")
AUTH = collections.namedtuple("Auth", "enabled host port path services default")

_ENVIRONMENT_MAPPER = {
    "rest.host": "API_GATEWAY_REST_HOST",
    "rest.port": "API_GATEWAY_REST_PORT",
    "rest.cors.enabled": "API_GATEWAY_REST_CORS_ENABLED",
    "rest.auth.enabled": "API_GATEWAY_REST_AUTH_ENABLED",
    "rest.auth.host": "API_GATEWAY_REST_AUTH_HOST",
    "rest.auth.port": "API_GATEWAY_REST_AUTH_PORT",
    "rest.auth.path": "API_GATEWAY_REST_AUTH_PATH",
    "database.dbname": "API_GATEWAY_DATABASE_NAME",
    "database.user": "API_GATEWAY_DATABASE_USER",
    "database.password": "API_GATEWAY_DATABASE_PASSWORD",
    "database.host": "API_GATEWAY_DATABASE_HOST",
    "database.port": "API_GATEWAY_DATABASE_PORT",
    "discovery.host": "API_GATEWAY_DISCOVERY_HOST",
    "discovery.port": "API_GATEWAY_DISCOVERY_PORT",
}

_PARAMETERIZED_MAPPER = {
    "rest.host": "api_gateway_rest_host",
    "rest.port": "api_gateway_rest_port",
    "rest.cors.enabled": "api_gateway_rest_cors_enabled",
    "rest.auth.enabled": "api_gateway_rest_auth_enabled",
    "rest.auth.host": "api_gateway_rest_auth_host",
    "rest.auth.port": "api_gateway_rest_auth_port",
    "rest.auth.path": "api_gateway_rest_auth_path",
    "database.database": "api_gateway_database_name",
    "database.user": "api_gateway_database_user",
    "database.password": "api_gateway_database_password",
    "database.host": "api_gateway_database_host",
    "database.port": "api_gateway_database_port",
    "discovery.host": "api_gateway_discovery_host",
    "discovery.port": "api_gateway_discovery_port",
}


class ApiGatewayConfig(abc.ABC):
    """Api Gateway config class."""

    __slots__ = ("_services", "_path", "_data", "_with_environment", "_parameterized")

    def __init__(self, path: t.Union[Path, str], with_environment: bool = True, **kwargs):
        if isinstance(path, Path):
            path = str(path)
        self._services = {}
        self._path = path
        self._load(path)
        self._with_environment = with_environment
        self._parameterized = kwargs

    @staticmethod
    def _file_exit(path: str) -> bool:
        if os.path.isfile(path):
            return True
        return False

    def _load(self, path):
        if self._file_exit(path):
            with open(path) as f:
                self._data = yaml.load(f, Loader=yaml.FullLoader)
        else:
            raise ApiGatewayConfigException(f"Check if this path: {path} is correct")

    def _get(self, key: str, **kwargs: t.Any) -> t.Any:
        if key in _PARAMETERIZED_MAPPER and _PARAMETERIZED_MAPPER[key] in self._parameterized:
            return self._parameterized[_PARAMETERIZED_MAPPER[key]]

        if self._with_environment and key in _ENVIRONMENT_MAPPER and _ENVIRONMENT_MAPPER[key] in os.environ:
            if os.environ[_ENVIRONMENT_MAPPER[key]] in ["true", "True", "false", "False"]:
                return bool(util.strtobool(os.environ[_ENVIRONMENT_MAPPER[key]]))
            return os.environ[_ENVIRONMENT_MAPPER[key]]

        def _fn(k: str, data: dict[str, t.Any]) -> t.Any:
            current, _, following = k.partition(".")

            part = data[current]
            if not following:
                return part

            return _fn(following, part)

        return _fn(key, self._data)

    @property
    def rest(self) -> REST:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return REST(
            host=self._get("rest.host"),
            port=int(self._get("rest.port")),
            cors=self._cors,
            auth=self._auth,
            admin=self._admin,
        )

    @property
    def _cors(self) -> CORS:
        """Get the cors config.

        :return: A ``CORS`` NamedTuple instance.
        """
        return CORS(enabled=self._get("rest.cors.enabled"))

    @property
    def _admin(self) -> REST_ADMIN:
        """Get the cors config.

        :return: A ``CORS`` NamedTuple instance.
        """
        return REST_ADMIN(username=self._get("rest.admin.username"), password=self._get("rest.admin.password"))

    @property
    def _auth(self) -> t.Optional[AUTH]:
        try:
            services = self._auth_services
            return AUTH(
                enabled=self._get("rest.auth.enabled"),
                host=self._get("rest.auth.host"),
                port=int(self._get("rest.auth.port")),
                path=self._get("rest.auth.path"),
                services=services,
                default=self._get("rest.auth.default"),
            )
        except KeyError:
            return None

    @property
    def _auth_services(self) -> list[AUTH_SERVICE]:
        info = self._get("rest.auth.services")
        services = [self._auth_service_entry(service) for service in info]
        return services

    @staticmethod
    def _auth_service_entry(service: dict[str, Any]) -> AUTH_SERVICE:
        return AUTH_SERVICE(name=service["name"],)

    @property
    def database(self) -> DATABASE:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return DATABASE(
            dbname=self._get("database.dbname"),
            user=self._get("database.user"),
            password=self._get("database.password"),
            host=self._get("database.host"),
            port=int(self._get("database.port")),
        )

    @property
    def discovery(self) -> DISCOVERY:
        """Get the rest config.

        :return: A ``REST`` NamedTuple instance.
        """
        return DISCOVERY(host=self._get("discovery.host"), port=int(self._get("discovery.port")))



from sqlalchemy import (
    or_,
)
from sqlalchemy.orm import (
    sessionmaker,
)

from .models import (
    AuthRule,
    AuthRuleDTO,
    AutzRule,
    AutzRuleDTO,
)


class Repository:
    def __init__(self, engine):
        self.engine = engine
        self.s = sessionmaker(bind=engine)
        self.session = self.s()

    def create_auth_rule(self, record: AuthRule):
        self.session.add(record)
        self.session.commit()
        return record.to_serializable_dict()

    def create_autz_rule(self, record: AutzRule):
        self.session.add(record)
        self.session.commit()
        return record.to_serializable_dict()

    def get_auth_rules(self):
        r = self.session.query(AuthRule).all()

        records = list()
        for record in r:
            records.append(AuthRuleDTO(record).__dict__)
        return records

    def get_autz_rules(self):
        r = self.session.query(AutzRule).all()

        records = list()
        for record in r:
            records.append(AutzRuleDTO(record).__dict__)
        return records

    def update_auth_rule(self, id: int, **kwargs):
        self.session.query(AuthRule).filter(AuthRule.id == id).update(kwargs)
        self.session.commit()

    def update_autz_rule(self, id: int, **kwargs):
        self.session.query(AutzRule).filter(AutzRule.id == id).update(kwargs)
        self.session.commit()

    def delete_auth_rule(self, id: int):
        self.session.query(AuthRule).filter(AuthRule.id == id).delete()
        self.session.commit()

    def delete_autz_rule(self, id: int):
        self.session.query(AutzRule).filter(AutzRule.id == id).delete()
        self.session.commit()

    def get_auth_rule_by_service(self, service: str):
        r = self.session.query(AuthRule).filter(or_(AuthRule.service == service, AuthRule.service == "*")).all()

        records = list()
        for record in r:
            records.append(AuthRuleDTO(record))
        return records

    def get_autz_rule_by_service(self, service: str):
        r = self.session.query(AutzRule).filter(or_(AutzRule.service == service, AutzRule.service == "*")).all()

        records = list()
        for record in r:
            records.append(AutzRuleDTO(record))
        return records



from ..database.models import (
    AutzRuleDTO,
)
from .urlmatch import (
    UrlMatch,
)


class AutzMatch(UrlMatch):
    @staticmethod
    def match(url: str, role: int, method: str, records: list[AutzRuleDTO]) -> bool:
        for record in records:
            if AutzMatch.urlmatch(record.rule, url):
                if record.roles is None:  # pragma: no cover
                    return True
                else:
                    if role in record.roles or "*" in record.roles:
                        return True

        return False



"""
urlmatch lets you easily check whether urls are on a domain or subdomain

match_pattern = 'http://*.example.com/*'

urlmatch(match_pattern, 'http://subdomain.example.com/') # True
urlmatch(match_pattern, 'http://sub.subdomain.example.com/') # True

urlmatch(match_pattern, 'https://example.com/') # False
urlmatch(match_pattern, 'http://bad.com/') # False

<url-pattern> := <scheme>://<host><path>
<scheme> := '*' | 'http' | 'https'
<host> := '*' | '*.' <any char except '/' and '*'>+
<path> := '/' <any chars>

"""
import re


class BadMatchPattern(Exception):
    """The Exception that's raised when a match pattern fails"""

    pass


class UrlMatch:
    @staticmethod
    def parse_match_pattern(pattern, path_required=True, fuzzy_scheme=False, http_auth_allowed=True):
        """
        Returns the regular expression for a given match pattern.

        Args:
            pattern: a `urlmatch` formatted match pattern
            path_required: a `bool` which dicates whether the match pattern
                must have path
            fuzzy_scheme: if this is true, then if the scheme is `*`, `http`,
                or `https`, it will match both `http` and `https`
            http_auth_allowed: if this is true, then URLs with http auth on the
                correct domain and path will be matched

        Returns:
            a regular expresion for the match pattern
        """
        pattern_regex = "(?:^"
        result = re.search(r"^(\*|https?):\/\/", pattern)
        if not result:
            raise BadMatchPattern("Invalid scheme: {}".format(pattern))
        elif result.group(1) == "*" or fuzzy_scheme:
            pattern_regex += "https?"
        else:
            pattern_regex += result.group(1)

        pattern_regex += "://"

        if http_auth_allowed:
            # add optional HTTP auth
            safe_characters = r"[^\/:.]"
            pattern_regex += r"(?:{safe}+(?:\:{safe}+)?@)?".format(safe=safe_characters)

        pattern = pattern[len(result.group(0)) :]  # noqa pylint: disable=whitespace

        domain_and_path_regex = r"^(?:\*|(\*\.)?([^\/*]+))"
        if path_required:
            domain_and_path_regex += r"(?=\/)"

        result = re.search(domain_and_path_regex, pattern)
        if not result:
            raise BadMatchPattern("Invalid domain or path: {}".format(pattern))

        pattern = pattern[len(result.group(0)) :]  # noqa pylint: disable=whitespace

        if result.group(0) == "*":
            pattern_regex += "[^/]+"
        else:
            if result.group(1):
                pattern_regex += "(?:[^/]+\\.)?"

            pattern_regex += re.escape(result.group(2))

        pattern_regex += ".*".join(map(re.escape, pattern.split("*")))
        pattern_regex += r"(\/.*)?$)"

        return pattern_regex

    @staticmethod
    def urlmatch(match_pattern, url, **kwargs):
        """
        Returns whether a given match pattern matches a url

        Args:
            match_pattern: a `urlmatch` formatted match pattern_regex
            url: a url
        """
        if isinstance(match_pattern, str):
            match_pattern = map(str.strip, match_pattern.split(","))

        regex = "({})".format("|".join(map(lambda x: UrlMatch.parse_match_pattern(x, **kwargs), match_pattern)))

        return bool(regex and re.search(regex, url))



import json
import logging
import secrets
from datetime import (
    datetime,
)
from typing import (
    Any,
    Optional,
)

from aiohttp import (
    ClientConnectorError,
    ClientResponse,
    ClientSession,
    web,
)
from yarl import (
    URL,
)

from minos.api_gateway.rest.database.models import (
    AuthRule,
    AutzRule,
)
from minos.api_gateway.rest.urlmatch.authmatch import (
    AuthMatch,
)

from .database.repository import (
    Repository,
)
from .urlmatch.autzmatch import (
    AutzMatch,
)

logger = logging.getLogger(__name__)


async def orchestrate(request: web.Request) -> web.Response:
    """ Orchestrate discovery and microservice call """
    discovery_host = request.app["config"].discovery.host
    discovery_port = request.app["config"].discovery.port

    verb = request.method
    url = f"/{request.match_info['endpoint']}"

    discovery_data = await discover(discovery_host, int(discovery_port), "/microservices", verb, url)

    auth = request.app["config"].rest.auth
    user = None
    if auth is not None and auth.enabled:
        if await check_authentication(
            request=request, service=request.url.parts[1], url=str(request.url), method=request.method
        ):
            response = await validate_token(request)
            user = json.loads(response)
            user = user["uuid"]

        if await check_authorization(
            request=request, service=request.url.parts[1], url=str(request.url), method=request.method
        ):
            response = await validate_token(request)
            data = json.loads(response)
            user = data["uuid"]
            role = data["role"]
            if not await is_authorized_role(
                request=request, role=role, service=request.url.parts[1], url=str(request.url), method=request.method
            ):
                return web.HTTPUnauthorized()

    microservice_response = await call(**discovery_data, original_req=request, user=user)
    return microservice_response


async def check_authentication(request: web.Request, service: str, url: str, method: str) -> bool:
    records = Repository(request.app["db_engine"]).get_auth_rule_by_service(service)
    return AuthMatch.match(url=url, method=method, records=records)


async def check_authorization(request: web.Request, service: str, url: str, method: str) -> bool:
    records = Repository(request.app["db_engine"]).get_autz_rule_by_service(service)
    return AuthMatch.match(url=url, method=method, records=records)


async def is_authorized_role(request: web.Request, role: int, service: str, url: str, method: str) -> bool:
    records = Repository(request.app["db_engine"]).get_autz_rule_by_service(service)
    return AutzMatch.match(url=url, role=role, method=method, records=records)


async def authentication_default(request: web.Request) -> web.Response:
    """ Orchestrate discovery and microservice call """
    auth_host = request.app["config"].rest.auth.host
    auth_port = request.app["config"].rest.auth.port
    auth_path = request.app["config"].rest.auth.path
    default_service = request.app["config"].rest.auth.default

    url = URL(f"http://{auth_host}:{auth_port}{auth_path}/{default_service}")

    return await authentication_call(request, url)


async def login_default(request: web.Request) -> web.Response:
    """ Orchestrate discovery and microservice call """
    auth_host = request.app["config"].rest.auth.host
    auth_port = request.app["config"].rest.auth.port
    auth_path = request.app["config"].rest.auth.path
    default_service = request.app["config"].rest.auth.default

    url = URL(f"http://{auth_host}:{auth_port}{auth_path}/{default_service}/login")

    return await authentication_call(request, url)


async def authentication(request: web.Request) -> web.Response:
    """ Orchestrate discovery and microservice call """

    auth_host = request.app["config"].rest.auth.host
    auth_port = request.app["config"].rest.auth.port

    url = URL(f"http://{auth_host}:{auth_port}{request.path}")

    return await authentication_call(request, url)


async def authentication_call(request: web.Request, url: URL) -> web.Response:
    """ Orchestrate discovery and microservice call """
    headers = request.headers.copy()
    data = await request.read()

    try:
        async with ClientSession() as session:
            async with session.request(headers=headers, method=request.method, url=url, data=data) as response:
                return await _clone_response(response)
    except ClientConnectorError:
        raise web.HTTPServiceUnavailable(text="The requested endpoint is not available.")


async def validate_token(request: web.Request):
    """ Orchestrate discovery and microservice call """
    auth_host = request.app["config"].rest.auth.host
    auth_port = request.app["config"].rest.auth.port
    auth_path = request.app["config"].rest.auth.path

    auth_url = URL(f"http://{auth_host}:{auth_port}{auth_path}/validate-token")

    headers = request.headers.copy()
    data = await request.read()

    try:
        async with ClientSession() as session:
            async with session.request(method="POST", url=auth_url, data=data, headers=headers) as response:
                resp = await _clone_response(response)

                if not response.ok:
                    raise web.HTTPUnauthorized(text="The given request does not have authorization to be forwarded.")
                return resp.text

    except ClientConnectorError:
        raise web.HTTPServiceUnavailable(text="The requested endpoint is not available.")


async def discover(host: str, port: int, path: str, verb: str, endpoint: str) -> dict[str, Any]:
    """Call discovery service and get microservice connection data.

    :param host: Discovery host name.
    :param port: Discovery port.
    :param path: Discovery path.
    :param verb: Endpoint Verb.
    :param endpoint: Endpoint url.
    :return: The response of the discovery.
    """

    url = URL.build(scheme="http", host=host, port=port, path=path, query={"verb": verb, "path": endpoint})
    try:
        async with ClientSession() as session:
            async with session.get(url=url) as response:
                if not response.ok:
                    if response.status == 404:
                        raise web.HTTPNotFound(text=f"The {endpoint!r} path is not available for {verb!r} method.")
                    raise web.HTTPBadGateway(text="The Discovery Service response is wrong.")

                data = await response.json()
    except ClientConnectorError:
        raise web.HTTPGatewayTimeout(text="The Discovery Service is not available.")

    data["port"] = int(data["port"])

    return data


# noinspection PyUnusedLocal
async def call(address: str, port: int, original_req: web.Request, user: Optional[str], **kwargs) -> web.Response:
    """Call microservice (redirect the original call)

    :param address: The ip of the microservices.
    :param port: The port of the microservice.
    :param original_req: The original request.
    :param kwargs: Additional named arguments.
    :param user: User that makes the request
    :return: The web response to be retrieved to the client.
    """

    headers = original_req.headers.copy()
    if user is not None:
        headers["X-User"] = user
    else:  # Enforce that the 'User' entry is only generated by the auth system.
        # noinspection PyTypeChecker
        headers.pop("X-User", None)

    url = original_req.url.with_scheme("http").with_host(address).with_port(port)
    method = original_req.method
    data = await original_req.read()

    logger.info(f"Redirecting {method!r} request to {url!r}...")

    try:
        async with ClientSession() as session:
            async with session.request(headers=headers, method=method, url=url, data=data) as response:
                return await _clone_response(response)
    except ClientConnectorError:
        raise web.HTTPServiceUnavailable(text="The requested endpoint is not available.")


# noinspection PyMethodMayBeStatic
async def _clone_response(response: ClientResponse) -> web.Response:
    return web.Response(
        body=await response.read(), status=response.status, reason=response.reason, headers=response.headers,
    )


class AdminHandler:
    @staticmethod
    async def login(request: web.Request) -> web.Response:
        """ Orchestrate discovery and microservice call """
        username = request.app["config"].rest.admin.username
        password = request.app["config"].rest.admin.password

        try:
            content = await request.json()

            if "user" not in content and "password" not in content:
                return web.json_response(
                    {"error": "Wrong data. Provide user and password."}, status=web.HTTPUnauthorized.status_code
                )

            if username == content["username"] and password == content["password"]:
                return web.json_response({"id": 1, "token": secrets.token_hex(20)})

            return web.json_response({"error": "Wrong username or password!."}, status=web.HTTPUnauthorized.status_code)
        except Exception:
            return web.json_response({"error": "Something went wrong!."}, status=web.HTTPUnauthorized.status_code)

    @staticmethod
    async def get_endpoints(request: web.Request) -> web.Response:
        discovery_host = request.app["config"].discovery.host
        discovery_port = request.app["config"].discovery.port

        url = URL.build(scheme="http", host=discovery_host, port=discovery_port, path="/endpoints")

        try:
            async with ClientSession() as session:
                async with session.get(url=url) as response:
                    return await _clone_response(response)
        except ClientConnectorError:
            return web.json_response(
                {"error": "The requested endpoint is not available."}, status=web.HTTPServiceUnavailable.status_code
            )

    @staticmethod
    async def get_roles(request: web.Request) -> web.Response:
        auth_host = request.app["config"].rest.auth.host
        auth_port = request.app["config"].rest.auth.port
        auth_path = request.app["config"].rest.auth.path

        url = URL.build(scheme="http", host=auth_host, port=auth_port, path=f"{auth_path}/roles")

        try:
            async with ClientSession() as session:
                async with session.get(url=url) as response:
                    return await _clone_response(response)
        except ClientConnectorError:
            return web.json_response(
                {"error": "The requested endpoint is not available."}, status=web.HTTPServiceUnavailable.status_code
            )

    @staticmethod
    async def get_rules(request: web.Request) -> web.Response:
        records = Repository(request.app["db_engine"]).get_auth_rules()
        return web.json_response(records)

    @staticmethod
    async def create_rule(request: web.Request) -> web.Response:
        try:
            content = await request.json()

            if "service" not in content and "rule" not in content and "methods" not in content:
                return web.json_response(
                    {"error": "Wrong data. Provide 'service', 'rule' and 'methods' parameters."},
                    status=web.HTTPBadRequest.status_code,
                )

            now = datetime.now()

            rule = AuthRule(
                service=content["service"],
                rule=content["rule"],
                methods=content["methods"],
                created_at=now,
                updated_at=now,
            )

            record = Repository(request.app["db_engine"]).create_auth_rule(rule)

            return web.json_response(record)
        except Exception as e:
            return web.json_response({"error": str(e)}, status=web.HTTPBadRequest.status_code)

    @staticmethod
    async def update_rule(request: web.Request) -> web.Response:
        try:
            id = int(request.url.name)
            content = await request.json()
            Repository(request.app["db_engine"]).update_auth_rule(id=id, **content)
            return web.json_response(status=web.HTTPOk.status_code)
        except Exception as e:
            return web.json_response({"error": str(e)}, status=web.HTTPBadRequest.status_code)

    @staticmethod
    async def update_autz_rule(request: web.Request) -> web.Response:
        try:
            id = int(request.url.name)
            content = await request.json()
            Repository(request.app["db_engine"]).update_autz_rule(id=id, **content)
            return web.json_response(status=web.HTTPOk.status_code)
        except Exception as e:
            return web.json_response({"error": str(e)}, status=web.HTTPBadRequest.status_code)

    @staticmethod
    async def delete_rule(request: web.Request) -> web.Response:
        try:
            id = int(request.url.name)
            Repository(request.app["db_engine"]).delete_auth_rule(id)
            return web.json_response(status=web.HTTPOk.status_code)
        except Exception as e:
            return web.json_response({"error": str(e)}, status=web.HTTPBadRequest.status_code)

    @staticmethod
    async def delete_autz_rule(request: web.Request) -> web.Response:
        try:
            id = int(request.url.name)
            Repository(request.app["db_engine"]).delete_autz_rule(id)
            return web.json_response(status=web.HTTPOk.status_code)
        except Exception as e:
            return web.json_response({"error": str(e)}, status=web.HTTPBadRequest.status_code)

    @staticmethod
    async def create_autz_rule(request: web.Request) -> web.Response:
        try:
            content = await request.json()

            if (
                "service" not in content
                and "rule" not in content
                and "roles" not in content
                and "methods" not in content
            ):
                return web.json_response(
                    {"error": "Wrong data. Provide 'service', 'rule', 'roles' and 'methods' parameters."},
                    status=web.HTTPBadRequest.status_code,
                )

            now = datetime.now()

            rule = AutzRule(
                service=content["service"],
                rule=content["rule"],
                roles=content["roles"],
                methods=content["methods"],
                created_at=now,
                updated_at=now,
            )

            record = Repository(request.app["db_engine"]).create_autz_rule(rule)

            return web.json_response(record)
        except Exception as e:
            return web.json_response({"error": str(e)}, status=web.HTTPBadRequest.status_code)

    @staticmethod
    async def get_autz_rules(request: web.Request) -> web.Response:
        records = Repository(request.app["db_engine"]).get_autz_rules()
        return web.json_response(records)



from sqlalchemy import (
    JSON,
    TIMESTAMP,
    Column,
    Integer,
    Sequence,
    String,
)
from sqlalchemy.ext.declarative import (
    declarative_base,
)

Base = declarative_base()


class AuthRule(Base):
    __tablename__ = "auth_rules"
    id = Column(Integer, Sequence("item_id_seq"), nullable=False)
    service = Column(String, primary_key=True, nullable=False)
    rule = Column(String, primary_key=True, nullable=False)
    methods = Column(JSON)
    created_at = Column(TIMESTAMP)
    updated_at = Column(TIMESTAMP)

    def __repr__(self):
        return (
            "<AuthRule(id='{}', service='{}', rule='{}',"
            "methods={}, created_at={}, updated_at={})>".format(  # pragma: no cover
                self.id, self.service, self.rule, self.methods, self.created_at, self.updated_at
            )
        )

    def to_serializable_dict(self):
        return {
            "id": self.id,
            "service": self.service,
            "rule": self.rule,
            "methods": self.methods,
            "created_at": str(self.created_at),
            "updated_at": str(self.updated_at),
        }


class AuthRuleDTO:
    id: int
    service: str
    rule: str
    methods: list
    created_at: str
    updated_at: str

    def __init__(self, model: AuthRule):
        self.id = model.id
        self.service = model.service
        self.rule = model.rule
        self.methods = model.methods
        self.created_at = str(model.created_at)
        self.updated_at = str(model.updated_at)


class AutzRule(Base):
    __tablename__ = "autz_rules"
    id = Column(Integer, Sequence("item_id_seq"), nullable=False, primary_key=True)
    service = Column(String, primary_key=True, nullable=False)
    rule = Column(String, primary_key=True, nullable=False)
    roles = Column(JSON)
    methods = Column(JSON)
    created_at = Column(TIMESTAMP)
    updated_at = Column(TIMESTAMP)

    def __repr__(self):
        return (
            "<AuthRule(id='{}', service='{}', rule='{}',"
            "methods={}, created_at={}, updated_at={})>".format(  # pragma: no cover
                self.id, self.service, self.roles, self.methods, self.created_at, self.updated_at
            )
        )

    def to_serializable_dict(self):
        return {
            "id": self.id,
            "service": self.service,
            "roles": self.roles,
            "methods": self.methods,
            "created_at": str(self.created_at),
            "updated_at": str(self.updated_at),
        }


class AutzRuleDTO:
    id: int
    service: str
    rule: str
    roles: list
    methods: list
    created_at: str
    updated_at: str

    def __init__(self, model: AutzRule):
        self.id = model.id
        self.rule = model.rule
        self.service = model.service
        self.roles = model.roles
        self.methods = model.methods
        self.created_at = str(model.created_at)
        self.updated_at = str(model.updated_at)



from ..database.models import (
    AuthRuleDTO,
)
from .urlmatch import (
    UrlMatch,
)


class AuthMatch(UrlMatch):
    @staticmethod
    def match(url: str, method: str, records: list[AuthRuleDTO]) -> bool:
        for record in records:
            if AuthMatch.urlmatch(record.rule, url):
                if record.methods is None:  # pragma: no cover
                    return True
                else:
                    if method in record.methods or "*" in record.methods:
                        return True
        return False



__version__ = "0.4.0"

from .config import (
    ApiGatewayConfig,
)
from .exceptions import (
    ApiGatewayConfigException,
    ApiGatewayException,
    NoTokenException,
)
from .launchers import (
    EntrypointLauncher,
)
from .service import (
    ApiGatewayRestService,
)
